/**************************************************************************************************
* NOTICE OF COPYRIGHT AND OWNERSHIP OF SOFTWARE:
*
* Copyright 2010 Flexera Software, Inc.  All Rights Reserved.  
*
* This computer program is the property of Flexera Software, Inc.  Any use, copy, publication, 
* distribution, display, modification, or transmission of this computer program in whole or in part
* in any form or by any means without the prior express written permission of Flexera Software, 
* Inc. is strictly prohibited.
* 
* Except when expressly provided by Flexera Software, Inc. in writing, possession of this computer 
* program shall not be construed to confer any license or rights under any of Flexera Software,
* Inc.'s intellectual property rights, whether by estoppel, implication, or otherwise.  
* 
* ALL COPIES OF THIS PROGRAM MUST DISPLAY THIS NOTICE OF COPYRIGHT AND OWNERSHIP IN FULL.
**************************************************************************************************/
/*
	This sample utility demonstrates Composite Transactions.

	If defines the function used for online transactions.

		comptranutil -transaction ... 

*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "FlxActCommon.h"
#include "FlxActError.h"

#include "FlxActTransaction.h"

#include "compTranUtilContext.h"
#include "compTranUtilSend.h"
#include "compTranUtilCreateRequest.h"
#include "compTranUtilManageRequests.h"
#include "compTranUtilResponse.h"

static uint32_t sCommStatusCallback(const void* pUserData, uint32_t statusOld, uint32_t statusNew);

typedef struct CommsCallbackData_struct
{
	FlxActBool			bDidConnect;
	FlxActBool			bDidSendRequest;
	FlxActBool			bDidReceiveResponse;
} CommsCallbackData;

static FlxActBool sSetComms(FlxCtuContext* pContext, const FlxCtuCommand* pCommand);
static void		  sSetCommsProxy(FlxCtuContext* pContext, const char* pszProxyDetails);
static void		  sSetCommsSsl(FlxCtuContext* pContext, const char* pszSslDetails);

/***************************************************************************************************
*	flxCtuDoCommandTransaction
*
* Perform a transaction with a server (send request, receive and process response).
***************************************************************************************************/
FlxActBool flxCtuDoCommandTransaction(FlxCtuContext* pContext, const FlxCtuCommand* pCommand)
{
	FlxActTranRequest	hTranRequest;
	CommsCallbackData	callbackData;
	FlxActBool			bIsNewRequest;

	/* Set the server comms attributes from the command. */
	if ( !sSetComms(pContext, pCommand) )
		return FLX_ACT_FALSE;

	/* Get a request, either new or existing. */
	bIsNewRequest = flxCtuContextHasCommand(pContext, CMD_NEW);
	if (bIsNewRequest)
	{
		/* A "new" command is present, get it. */
		FlxCtuCommand* pCommandNew;
		if ( !flxCtuContextGetCommandByName(pContext, CMD_NEW, (const FlxCtuCommand**)&pCommandNew))
			FLX_CTU_ASSERT(FLX_ACT_FALSE);	/* Know command exists so should not happen. */

		/* Create and save a new request, or match a stored one that has identical attributes. */
		if ( !flxCtuRequestCreate(pContext, pCommandNew, &hTranRequest) )
			return FLX_ACT_FALSE;
	}
	else if (flxCtuContextHasCommand(pContext, CMD_STORED))
	{	
		/* Use stored request. */
		const FlxCtuCommand* pCommandStored;
		if ( !flxCtuContextGetCommandByName(pContext, CMD_STORED, &pCommandStored) )
			FLX_CTU_EXIT_UNEXPECTED_API_ERROR(pContext);

		if ( !flxCtuLoadStoredRequest(pContext, pCommandStored) )
			return FLX_ACT_FALSE;
	}
	else
	{	/* No request specified. */
		flxCtuPrintError("No request specified for %s (use -new or -stored)",
						 pCommand->pszNameForMessages);
		return FLX_ACT_FALSE;
	}

	/* Initialise the data that will be set by the comms status callback function. */
	callbackData.bDidConnect		 = FLX_ACT_FALSE;
	callbackData.bDidSendRequest	 = FLX_ACT_FALSE;
	callbackData.bDidReceiveResponse = FLX_ACT_FALSE;

	/* Send the request to the server and process the response it sends back. */
	if (flxActTransactionSend(pContext->hTransaction, sCommStatusCallback, &callbackData))
	{
		flxCtuResponseOnProcessSuccess(pContext);
	}
	else
	{
		if (callbackData.bDidReceiveResponse)
		{	/* Response received but error when processing it. */
			flxCtuResponseOnProcessError(pContext);
		}
		else 
		{	/* Comms unsuccessful. */
			FlxActTranResult	 result  = flxActTransactionGetResult();
			const char* serverUri;

			flxCtuContextPrintLastApiError(pContext);

			if ( !flxActTransactionGetCommsAttribute(pContext->hTransaction, 
													 FLX_ACT_TRAN_SVR_URI, 
													 &serverUri))
				FLX_CTU_EXIT_UNEXPECTED_API_ERROR(pContext);

			if (callbackData.bDidSendRequest)
			{	/* Request sent but no response received. Don't know whether the server processed 
				   it or not so leave the request in trusted storage for re-sending. */
				flxCtuPrintError("No response from server %s\n", serverUri);
			}
			else
			{	/* Unable to connect/send request - the server can't have received it so it can be
				   canceled, re-enabling any FRs that were being returned. If/when the user retries,
				   a new request will be created. 
				   We only do this if the request was created for this transaction - if an existing
				   stored request was used we leave it for the user to manage. */
				if ( bIsNewRequest)
				{
					if ( !flxActTransactionDeleteStoredRequest(pContext->hTransaction) )
					{
						flxCtuPrintError("Unable to delete request\n");
						flxCtuContextPrintLastApiError(pContext);
					}
				}
				if (callbackData.bDidConnect)
				{
					flxCtuPrintError("Unable to send request to server %s (result %u)\n", 
									 serverUri,
									 result);
				}
				else if (result == FLX_ACT_TRAN_ERR_COMMS_INIT)
				{
					flxCtuPrintError("Unable to initialise comms (fnpCommsSoap shared object missing?)\n");
				}
				else
				{
					flxCtuPrintError("Unable to connect to server %s\n", serverUri);
				}
			}
			pContext->exitCode = flxCtuExitCommsError;
		}
		return FLX_ACT_FALSE;
	}

	/* Transaction completed successfully - delete the request. */
	if ( !flxActTransactionDeleteStoredRequest(pContext->hTransaction) )
	{
		flxCtuPrintError("Unable to delete request\n");
		flxCtuContextPrintLastApiError(pContext);
		return FLX_ACT_FALSE;
	}
	return FLX_ACT_TRUE;
}

/***************************************************************************************************
*	sCommStatusCallback
*
* This is called from within flxActTransactionSend when each stage of the transaction is completed 
* and the status changes.
***************************************************************************************************/
static uint32_t sCommStatusCallback(const void* pUserData, uint32_t statusOld, uint32_t statusNew)
{
	const char* const COMMS_STATUS_STRINGS[] = {
		"Error",
		"Success",
		"Not set",
		"Canceled by caller",
		"Creating request",
		"Request created",
		"Context created",
		"Connected to remote server",
		"Request Sent",
		"Polling for response",
		"Waiting for response",
		"Done"
	};
	CommsCallbackData* pCallbackData = (CommsCallbackData*)pUserData;

	printf("Status: %d, %s\n", statusNew, COMMS_STATUS_STRINGS[statusNew]);

	if (statusNew == flxCommsStatusConnected)
	{
		/* Successfully connected to the server - tell the caller through the user data. */
		pCallbackData->bDidConnect = FLX_ACT_TRUE;
	}
	else if (statusNew == flxCommsStatusRequestSent)
	{
		/* Successfully sent request to the server - tell the caller through the user data. */
		pCallbackData->bDidSendRequest = FLX_ACT_TRUE;
	}
	else if (statusNew == flxCommsStatusDone)
	{
		/* Received a response from the server - tell the caller through the user data. */
		pCallbackData->bDidReceiveResponse = FLX_ACT_TRUE;
	}

	return flxCallbackReturnContinue;
}

/***************************************************************************************************
 *	sSetComms
 *
 * Set the server communications attributes from the command.
 ***************************************************************************************************/
static FlxActBool sSetComms(FlxCtuContext* pContext, const FlxCtuCommand* pCommand)
{
	const char* pszServerUri = flxCtuCommandGetValue(pCommand); /* Mandatory. */
	/* Set the server URI. */
	if ( !flxActTransactionSetCommsAttribute(pContext->hTransaction, 
											 FLX_ACT_TRAN_SVR_URI, 
											 pszServerUri) )
	{
		flxCtuContextPrintLastApiError(pContext);
		return FLX_ACT_FALSE;
	}
	flxCtuPrintVerbose("Server URI:     %s\n", pszServerUri);

	/* Set any proxy attributes. */
	if ( flxCtuCommandHasAttribute(pCommand, CMD_ATTR_PROXY) )
		sSetCommsProxy(pContext, flxCtuCommandGetAttribute(pCommand, CMD_ATTR_PROXY));
	
	/* Set any SSL attributes. */
	if ( flxCtuCommandHasAttribute(pCommand, CMD_ATTR_SSL) )
		sSetCommsSsl(pContext, flxCtuCommandGetAttribute(pCommand, CMD_ATTR_SSL));
	
	/* Set any timeout. */
	if ( flxCtuCommandHasAttribute(pCommand, CMD_ATTR_TIMEOUT) )
	{
		const char* pszTimeout = flxCtuCommandGetAttribute(pCommand, CMD_ATTR_TIMEOUT);
		/* Set the attribute; this call should not fail given valid parameters. */
		if ( !flxActTransactionSetCommsAttribute(pContext->hTransaction,
												 FLX_ACT_TRAN_SVR_TIMEOUT,
												 pszTimeout) )
 			FLX_CTU_EXIT_UNEXPECTED_API_ERROR(pContext);

		flxCtuPrintVerbose("Timeout:        %s\n", pszTimeout);
	}

	/* Set any poll interval. */
	if ( flxCtuCommandHasAttribute(pCommand, CMD_ATTR_POLL) )
	{
		const char* pszPollInterval = flxCtuCommandGetAttribute(pCommand, CMD_ATTR_POLL);
		/* Set the attribute; this call should not fail given valid parameters. */
		if ( !flxActTransactionSetCommsAttribute(pContext->hTransaction,
												 FLX_ACT_TRAN_SVR_POLL_INTERVAL,
												 pszPollInterval) )
			FLX_CTU_EXIT_UNEXPECTED_API_ERROR(pContext);

		flxCtuPrintVerbose("Poll interval:  %s\n", pszPollInterval);
	}

	return FLX_ACT_TRUE;
}

/***************************************************************************************************
 *	sSetCommsProxy
 *
 * Set the server communications proxy attributes.  The details are in the form
 * "host port userId password".
 ***************************************************************************************************/
static void sSetCommsProxy(FlxCtuContext* pContext, const char* pszProxyDetails)
{
	/* Parse the details and duplicate each space-delimited item (empty string if missing). */
	const char* pszHost		= flxCtuStrDupDelim(&pszProxyDetails, ' ');
	const char* pszPort		= flxCtuStrDupDelim(&pszProxyDetails, ' ');
	const char* pszUserId	= flxCtuStrDupDelim(&pszProxyDetails, ' ');
	const char* pszPassword = flxCtuStrDupDelim(&pszProxyDetails, ' ');

	/* Set the attributes; these calls should not fail given valid parameters. */
	if (strlen(pszHost) > 0)
		if ( !flxActTransactionSetCommsAttribute(pContext->hTransaction,
												 FLX_ACT_TRAN_SVR_PROXY_HOST,
												 pszHost))
			FLX_CTU_EXIT_UNEXPECTED_API_ERROR(pContext);

	if (strlen(pszPort) > 0)
		if ( !flxActTransactionSetCommsAttribute(pContext->hTransaction,
												 FLX_ACT_TRAN_SVR_PROXY_PORT,
												 pszPort))
			FLX_CTU_EXIT_UNEXPECTED_API_ERROR(pContext);

	if (strlen(pszUserId) > 0)
		if ( !flxActTransactionSetCommsAttribute(pContext->hTransaction,
												 FLX_ACT_TRAN_SVR_PROXY_USER,
												 pszUserId))
			FLX_CTU_EXIT_UNEXPECTED_API_ERROR(pContext);

	if (strlen(pszPassword) > 0)
		if ( !flxActTransactionSetCommsAttribute(pContext->hTransaction,
												 FLX_ACT_TRAN_SVR_PROXY_PASSWORD,
												 pszPassword))
			FLX_CTU_EXIT_UNEXPECTED_API_ERROR(pContext);

	flxCtuPrintVerbose("Proxy settings: host=%s port=%s username=%s password=%s\n",
					   pszHost,
					   pszPort,
					   pszUserId,
					   pszPassword);

	/* Free the duplicated strings. */
	flxCtuFree(pszHost);
	flxCtuFree(pszPort);
	flxCtuFree(pszUserId);
	flxCtuFree(pszPassword);
}

/***************************************************************************************************
 *	sSetCommsSsl
 *
 * Set the server communications SSL attributes.  The details are in the form
 * "CaCert CaPath".
 ***************************************************************************************************/
static void sSetCommsSsl(FlxCtuContext* pContext, const char* pszSslDetails)
{
	/* Parse the details and duplicate each space-delimited item (empty string if missing). */
	const char* pszCaCert	= flxCtuStrDupDelim(&pszSslDetails, ' ');
	const char* pszCaPath	= flxCtuStrDupDelim(&pszSslDetails, ' ');

	/* Set the attributes; these calls should not fail given valid parameters. */
	if (strlen(pszCaCert) > 0)
		if ( !flxActTransactionSetCommsAttribute(pContext->hTransaction,
														   FLX_ACT_TRAN_SVR_SSL_CACERT,
														   pszCaCert))
			FLX_CTU_EXIT_UNEXPECTED_API_ERROR(pContext);

	if (strlen(pszCaPath) > 0)
		if ( !flxActTransactionSetCommsAttribute(pContext->hTransaction,
												 FLX_ACT_TRAN_SVR_SSL_CAPATH,
												 pszCaPath))
			FLX_CTU_EXIT_UNEXPECTED_API_ERROR(pContext);

	flxCtuPrintVerbose("SSL settings:   CaCert=%s CaPath=%s\n",
					   pszCaCert,
					   pszCaPath);

	/* Free the duplicated strings. */
	flxCtuFree(pszCaCert);
	flxCtuFree(pszCaPath);
}

